use std::cell::RefCell;
use std::rc::Rc;

mod msg;
mod node;

mod crypto;
mod io;
mod helper;
pub mod service;

pub mod prelude {
    pub use crate::msg::Message;
    pub use crate::service::{Service, ServiceAPI};
}

use prelude::*;

use io::FleckIO;

pub struct Fleck {
    io: Rc<FleckIO>,
    services: RefCell<service::ServiceStack>,
    message_registry: msg::MessageRegistry,

    queued_packets: RefCell<Vec<io::Packet>>,
}

impl Fleck {
    pub fn new() -> Rc<Self> {
        let res = Rc::new(Self {
            io: Rc::new(FleckIO::default()),
            services: RefCell::new(service::ServiceStack::new()),
            message_registry: msg::MessageRegistry::new(),
            queued_packets: RefCell::new(vec![]),
        });

        res.register_core_services();

        res
    }

    fn register_core_services(&self) {
        let mut svcs = self.services.borrow_mut();
        // Node registration
        svcs.give_service(service::core::Nodes::default());
        // Local discovery
        svcs.give_service(service::core::LocalDiscovery::default());
        // Parsing incoming packets
        svcs.give_service(service::core::ParsePacket::default());
        // Actually sending packets
        svcs.give_service(service::core::SendPacket::default());
    }

    pub fn run(&self) {
        self.io.poll(self);
    }

    fn send_queued_packets(&self) {
        let svcs = self.services.borrow();
        for mut pkt in self.queued_packets.borrow_mut().drain(..) {
            svcs.process_outgoing(self, &mut pkt);
        }
    }
}

impl ServiceAPI for Fleck {
    fn raw_io(&self) -> &FleckIO {
        &self.io
    }
    fn send_packet(&self, pkt: io::Packet) {
        self.queued_packets.borrow_mut().push(pkt);
    }
    fn services(&self) -> std::cell::Ref<service::ServiceStack> {
        self.services.borrow()
    }
}

impl io::IOFeedback for Fleck {
    fn packet(&self, mut packet: io::Packet) {
        self.services.borrow().process_incoming(self, &mut packet);
        self.send_queued_packets();
    }

    fn minor_tick(&self) {
        self.services.borrow().process_minor_tick(self);
        self.send_queued_packets();
    }

    fn major_tick(&self) {
        self.services.borrow().process_major_tick(self);
        self.send_queued_packets();
    }
}
