use std::{
    any::Any,
    hash::{Hash, Hasher},
};

use serde::{de::DeserializeOwned, Deserialize, Serialize};
use std::collections::HashMap;

const MESSAGE_MAGIC: u64 = 0x1234123412341234;

fn message_type<M: MessageParams>() -> u16 {
    let mut hasher = std::collections::hash_map::DefaultHasher::new();
    M::NAME.hash(&mut hasher);
    (hasher.finish() & 0xffff) as u16
}

#[derive(Serialize, Deserialize)]
pub struct Message {
    magic: u64,
    ty: u16,
    data: Vec<u8>,
}

pub struct MessageBuilder {}

impl Message {
    pub fn build<M: MessageParams + Serialize>(from: M) -> Self {
        Self {
            magic: MESSAGE_MAGIC,
            ty: message_type::<M>(),
            data: vec![]
        }
    }
}

type Deserializer = dyn Fn(&[u8]) -> Option<Box<dyn std::any::Any>>;
pub struct MessageRegistry {
    deser: HashMap<u16, Box<Deserializer>>,
}

pub trait MessageParams {
    const NAME: &'static str;
    const ENCRYPTED: bool = true;
}

impl MessageRegistry {
    pub(crate) fn new() -> Self {
        Self {
            deser: Default::default(),
        }
    }

    pub fn add_message_type<MT: MessageParams + Serialize + DeserializeOwned>(&mut self) {
        let derived_typeid = message_type::<MT>();
        self.deser.insert(
            derived_typeid,
            Box::new(|data| {
                let boxed: Box<dyn Any> = Box::new(bincode::deserialize(data).ok()?);
                Some(boxed)
            }),
        );
    }

    pub(crate) fn deserialize(&self, data: &[u8]) -> Option<Box<dyn std::any::Any>> {
        let msg_type = u16::from_le_bytes(data[0..1].try_into().unwrap());
        match self.deser.get(&msg_type) {
            Some(f) => f(&data[2..]),
            None => None,
        }
    }
}
