use std::any::TypeId;
use std::cell::RefCell;
use std::collections::HashMap;
use std::ops::DerefMut;
use std::ops::Deref;
use std::{borrow::BorrowMut, rc::Rc};

use crate::io::{FleckIO, Packet};

mod lowlevel;
mod priority;
mod nodes;

pub use priority::{Never, ServicePriority as Priority};

use self::priority::AbstractServicePriority;

pub mod order {
    pub use super::priority::Never;

    pub struct FirstTag;
    pub type First = super::priority::Fixpoint<FirstTag>;

    pub struct PreprocessingTag;
    pub type Preprocessing = super::priority::Fixpoint<PreprocessingTag>;

    pub struct ProcessingTag;
    pub type Processing = super::priority::Fixpoint<ProcessingTag>;

    pub struct PostprocessingTag;
    pub type Postprocessing = super::priority::Fixpoint<PostprocessingTag>;

    pub struct LastTag;
    pub type Last = super::priority::Fixpoint<FirstTag>;
}

pub mod core {
    pub(crate) use super::lowlevel::{LocalDiscovery, SendPacket, ParsePacket};

    pub use super::nodes::Nodes;
}

pub trait ServiceAPI {
    fn raw_io(&self) -> &FleckIO;
    fn send_packet(&self, pkt: Packet);
    fn services(&self) -> std::cell::Ref<ServiceStack>;
}

pub trait Service : std::any::Any + crate::helper::AsAny {
    fn process_incoming(&mut self, _api: &dyn ServiceAPI, _packet: &mut Packet) {}
    fn process_outgoing(&mut self, _api: &dyn ServiceAPI, _packet: &mut Packet) {}
    fn process_minor_tick(&mut self, _api: &dyn ServiceAPI) {}
    fn process_major_tick(&mut self, _api: &dyn ServiceAPI) {}

    fn register_msgs(&self, _registry: &mut crate::msg::MessageRegistry) {}
}

pub trait IncomingPriority {
    type Priority: priority::AbstractServicePriority;
}

pub trait OutgoingPriority {
    type Priority: priority::AbstractServicePriority;
}

#[macro_export]
macro_rules! service_priority {
    ($what:ty, in: $i:ty, out: $o:ty) => {
        impl fleck::service::IncomingPriority for $what {
            type Priority = $i;
        }
        impl fleck::service::OutgoingPriority for $what {
            type Priority = $o;
        }
    };
    ($what:ty, in: $i:ty) => {
        impl fleck::service::IncomingPriority for $what {
            type Priority = $i;
        }
        impl fleck::service::OutgoingPriority for $what {
            type Priority = fleck::service::order::Never;
        }
    };
    ($what:ty, out: $o:ty) => {
        impl fleck::service::IncomingPriority for $what {
            type Priority = fleck::service::order::Never;
        }
        impl fleck::service::OutgoingPriority for $what {
            type Priority = $o;
        }
    };
    ($what:ty) => {
        impl fleck::service::IncomingPriority for $what {
            type Priority = fleck::service::order::Never;
        }
        impl fleck::service::OutgoingPriority for $what {
            type Priority = fleck::service::order::Never;
        }
    };
}

#[derive(Default)]
pub struct ServiceStack {
    services: Vec<Rc<RefCell<dyn Service>>>,

    service_map: HashMap<TypeId, Rc<RefCell<dyn Service>>>,

    incoming_order: priority::TotalOrder<Rc<RefCell<dyn Service>>>,
    outgoing_order: priority::TotalOrder<Rc<RefCell<dyn Service>>>,

    incoming_cache: RefCell<Option<Vec<Rc<RefCell<dyn Service>>>>>,
    outgoing_cache: RefCell<Option<Vec<Rc<RefCell<dyn Service>>>>>,
}

impl ServiceStack {
    pub(crate) fn new() -> Self {
        Self::default()
    }

    pub fn give_service<S: 'static + Service + IncomingPriority + OutgoingPriority>(
        &mut self,
        srv: S,
    ) {
        let srv = Rc::new(RefCell::new(srv));

        self.incoming_cache = RefCell::new(None);
        self.outgoing_cache = RefCell::new(None);

        if !<<S as IncomingPriority>::Priority as AbstractServicePriority>::NEVER {
            self.incoming_order
                .add_priority::<<S as IncomingPriority>::Priority>(srv.clone());
        }
        if !<<S as OutgoingPriority>::Priority as AbstractServicePriority>::NEVER {
            self.outgoing_order
                .add_priority::<<S as OutgoingPriority>::Priority>(srv.clone());
        }
        self.service_map.insert(TypeId::of::<S>(), srv.clone());
        self.services.push(srv);
    }

    pub fn with_service<S: Service + 'static, R, F: FnOnce(&S) -> R>(&self, f: F) -> R {
        let id = TypeId::of::<S>();

        self.service_map
            .get(&id)
            .expect("asked for service that doesn't exist!")
            .borrow()
            .deref()
            .as_any()
            .downcast_ref()
            .map(f)
            .unwrap()
    }
}

impl ServiceStack {
    pub(crate) fn process_incoming(&self, api: &dyn ServiceAPI, packet: &mut Packet) {
        let mut cache = self.incoming_cache.borrow_mut();
        let services = cache.get_or_insert_with(|| self.incoming_order.clone().order());

        for srv in services {
            if packet.is_clear() { break }
            srv.as_ref()
                .borrow_mut()
                .deref_mut()
                .process_incoming(api, packet);
        }
    }

    pub(crate) fn process_outgoing(&self, api: &dyn ServiceAPI, packet: &mut Packet) {
        let mut cache = self.outgoing_cache.borrow_mut();
        let services = cache.get_or_insert_with(|| self.outgoing_order.clone().order());

        for srv in services {
            if packet.is_clear() { break }
            srv.as_ref()
                .borrow_mut()
                .deref_mut()
                .process_outgoing(api, packet);
        }
    }

    pub(crate) fn process_minor_tick(&self, api: &dyn ServiceAPI) {
        for srv in &self.services {
            srv.as_ref()
                .borrow_mut()
                .deref_mut()
                .process_minor_tick(api);
        }
    }

    pub(crate) fn process_major_tick(&self, api: &dyn ServiceAPI) {
        for srv in &self.services {
            srv.as_ref()
                .borrow_mut()
                .deref_mut()
                .process_major_tick(api);
        }
    }
}
