use std::{
    any::TypeId,
    collections::{HashMap, HashSet},
};

pub trait AbstractServicePriority: 'static + Default {
    /// What this priority comes strictly after
    type After: AbstractServicePriority;
    /// What this priority comes strictly before
    type Before: AbstractServicePriority;

    /// Is this a fully-abstract node that should be merged during processing?
    const SKIP: bool = false;

    /// Is this a Never sentinel type?
    const NEVER: bool = false;
}

#[derive(Default)]
pub struct ServicePriority<After: AbstractServicePriority, Before: AbstractServicePriority> {
    _ghost: std::marker::PhantomData<(After, Before)>,
}

impl<After: AbstractServicePriority, Before: AbstractServicePriority> AbstractServicePriority
    for ServicePriority<After, Before>
{
    type After = After;
    type Before = Before;

    const SKIP: bool = false;
}

#[derive(Default)]
pub struct Never;

pub struct MaxTag;
pub struct MinTag;

#[allow(unused)]
pub type MinPriority = Fixpoint<MinTag>;
#[allow(unused)]
pub type MaxPriority = Fixpoint<MaxTag>;

pub struct Fixpoint<Tag: 'static> {
    _ghost: std::marker::PhantomData<Tag>,
}

impl AbstractServicePriority for Never {
    type After = Self;
    type Before = Self;

    const NEVER: bool = true;
}

impl<Tag: 'static> Default for Fixpoint<Tag> {
    fn default() -> Self {
        Self {
            _ghost: std::marker::PhantomData,
        }
    }
}

impl<Tag: 'static> AbstractServicePriority for Fixpoint<Tag> {
    type After = Self;
    type Before = Self;

    const SKIP: bool = true;
}

#[derive(Clone)]
pub(crate) struct TotalOrder<Assoc: Clone> {
    // directional adjacency-map of happens-after relationships
    nodes: HashMap<TypeId, HashSet<TypeId>>,
    associated: HashMap<TypeId, Vec<Option<Assoc>>>,
    skip: HashSet<TypeId>,
}

impl<Assoc: Clone> Default for TotalOrder<Assoc> {
    fn default() -> Self {
        Self {
            nodes: Default::default(),
            associated: Default::default(),
            skip: Default::default(),
        }
    }
}

type VisitedSet = HashSet<TypeId>;
impl<Assoc: Clone> TotalOrder<Assoc> {
    fn walk_priorities<SP: AbstractServicePriority>(
        &mut self,
        visited: &mut VisitedSet,
        assoc: &Option<Assoc>,
    ) {
        let tid = TypeId::of::<SP>();
        if visited.contains(&tid) {
            return;
        }
        visited.insert(tid);
        if SP::SKIP {
            self.skip.insert(tid);
        }

        // add happens-after link
        let after_id = TypeId::of::<SP::After>();
        let before_id = TypeId::of::<SP::Before>();
        self.nodes
            .entry(tid)
            .or_insert_with(Default::default)
            .insert(after_id);
        self.nodes
            .entry(before_id)
            .or_insert_with(Default::default)
            .insert(tid);

        // if we're not at an internal node, continue
        if self
            .associated
            .entry(tid)
            .or_insert(Vec::new())
            .iter()
            .all(Option::is_some)
        {
            self.walk_priorities::<SP::Before>(visited, assoc);
            self.walk_priorities::<SP::After>(visited, assoc);
        }
    }

    fn add_priority_internal<SP: AbstractServicePriority>(&mut self, assoc: Option<Assoc>) {
        let tid = TypeId::of::<SP>();
        self.associated
            .entry(tid)
            .or_insert_with(Vec::new)
            .push(assoc.clone());

        let mut visited = VisitedSet::default();
        self.walk_priorities::<SP>(&mut visited, &assoc);
    }

    pub(crate) fn add_priority<SP: AbstractServicePriority>(&mut self, assoc: Assoc) {
        self.add_priority_internal::<SP>(Some(assoc));
    }
}

impl<Assoc: Clone> TotalOrder<Assoc> {
    pub(crate) fn order(mut self) -> Vec<Assoc> {
        let mut tsort = topological_sort::TopologicalSort::<TypeId>::new();

        for node in self.nodes {
            tsort.insert(node.0);
            for adj in node.1 {
                tsort.insert(adj);

                if adj != node.0 || !self.skip.contains(&adj) {
                    tsort.add_dependency(adj, node.0);
                }
            }
        }

        let mut tid_order = vec![];
        while tsort.pop().map(|tid| tid_order.push(tid)).is_some() {}
        // we expect there to be two items left in the tsort: MaxPriority and Never
        if !tsort.is_empty() {
            panic!("Circular dependency detected! tsort: {:?}", tsort);
        }

        tid_order
            .iter()
            .flat_map(|tid| self.associated.remove(tid))
            .flatten()
            .map(|a| a.unwrap())
            .collect()
    }
}

#[cfg(test)]
mod order_tests {
    use super::{AbstractServicePriority, ServicePriority, TotalOrder};
    use super::{MaxPriority, MinPriority};

    #[test]
    fn single_priority_test() {
        let mut order = TotalOrder::default();
        order.add_priority::<ServicePriority<MinPriority, MaxPriority>>("simple priority");
        // println!("nodes: {:?}", order.nodes);
        let result = order.order();
        // println!("ordering: {:?}", result);
        assert_eq!(result, vec!["simple priority"]);
    }

    #[test]
    fn two_priority_test() {
        let mut order = TotalOrder::default();
        type P1 = ServicePriority<MinPriority, MaxPriority>;
        type P2 = ServicePriority<P1, MaxPriority>;
        order.add_priority::<P1>("P1");
        order.add_priority::<P2>("P2");
        let result = order.order();
        assert_eq!(result, vec!["P1", "P2"]);
    }
}
